"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const isNothing = (node) => {
    if (!node) {
        return true;
    }
    switch (node.type) {
        case "Literal":
            return [null, undefined, false, ""].includes(node.value);
        case "JSXFragment":
            return !node.children || node.children.every(isNothing);
        default:
            return false;
    }
};
const getLineLength = (loc) => loc.end.line - loc.start.line + 1;
const rule = {
    meta: {
        type: "problem",
        docs: {
            recommended: "error",
            description: "Disallow early returns in components. Solid components only run once, and so conditionals should be inside JSX.",
            url: "https://github.com/solidjs-community/eslint-plugin-solid/blob/main/docs/components-return-once.md",
        },
        fixable: "code",
        schema: [],
        messages: {
            noEarlyReturn: "Solid components run once, so an early return breaks reactivity. Move the condition inside a JSX element, such as a fragment or <Show />.",
            noConditionalReturn: "Solid components run once, so a conditional return breaks reactivity. Move the condition inside a JSX element, such as a fragment or <Show />.",
        },
    },
    create(context) {
        const functionStack = [];
        const putIntoJSX = (node) => {
            const text = context.getSourceCode().getText(node);
            return node.type === "JSXElement" || node.type === "JSXFragment" ? text : `{${text}}`;
        };
        const currentFunction = () => functionStack[functionStack.length - 1];
        const onFunctionEnter = (node) => {
            let lastReturn;
            if (node.body.type === "BlockStatement") {
                const { length } = node.body.body;
                const last = length && node.body.body[length - 1];
                if (last && last.type === "ReturnStatement") {
                    lastReturn = last;
                }
            }
            functionStack.push({ isComponent: false, lastReturn, earlyReturns: [] });
        };
        const onFunctionExit = (node) => {
            var _a, _b, _c, _d, _e, _f;
            if ((node.type === "FunctionDeclaration" && ((_b = (_a = node.id) === null || _a === void 0 ? void 0 : _a.name) === null || _b === void 0 ? void 0 : _b.match(/^[a-z]/))) ||
                ((_c = node.parent) === null || _c === void 0 ? void 0 : _c.type) === "JSXExpressionContainer" ||
                (((_d = node.parent) === null || _d === void 0 ? void 0 : _d.type) === "CallExpression" &&
                    node.parent.arguments.some((n) => n === node) &&
                    !((_e = node.parent.callee.name) === null || _e === void 0 ? void 0 : _e.match(/^[A-Z]/)))) {
                currentFunction().isComponent = false;
            }
            if (currentFunction().isComponent) {
                currentFunction().earlyReturns.forEach((earlyReturn) => {
                    context.report({
                        node: earlyReturn,
                        messageId: "noEarlyReturn",
                    });
                });
                const argument = (_f = currentFunction().lastReturn) === null || _f === void 0 ? void 0 : _f.argument;
                if ((argument === null || argument === void 0 ? void 0 : argument.type) === "ConditionalExpression") {
                    const sourceCode = context.getSourceCode();
                    context.report({
                        node: argument.parent,
                        messageId: "noConditionalReturn",
                        fix: (fixer) => {
                            const { test, consequent, alternate } = argument;
                            const conditions = [{ test, consequent }];
                            let fallback = alternate;
                            while (fallback.type === "ConditionalExpression") {
                                conditions.push({ test: fallback.test, consequent: fallback.consequent });
                                fallback = fallback.alternate;
                            }
                            if (conditions.length >= 2) {
                                const fallbackStr = !isNothing(fallback)
                                    ? ` fallback={${sourceCode.getText(fallback)}}`
                                    : "";
                                return fixer.replaceText(argument, `<Switch${fallbackStr}>\n${conditions
                                    .map(({ test, consequent }) => `<Match when={${sourceCode.getText(test)}}>${putIntoJSX(consequent)}</Match>`)
                                    .join("\n")}\n</Switch>`);
                            }
                            if (isNothing(consequent)) {
                                return fixer.replaceText(argument, `<Show when={!(${sourceCode.getText(test)})}>${putIntoJSX(alternate)}</Show>`);
                            }
                            if (isNothing(fallback) ||
                                getLineLength(consequent.loc) >= getLineLength(alternate.loc) * 1.5) {
                                const fallbackStr = !isNothing(fallback)
                                    ? ` fallback={${sourceCode.getText(fallback)}}`
                                    : "";
                                return fixer.replaceText(argument, `<Show when={${sourceCode.getText(test)}}${fallbackStr}>${putIntoJSX(consequent)}</Show>`);
                            }
                            return fixer.replaceText(argument, `<>${putIntoJSX(argument)}</>`);
                        },
                    });
                }
                else if ((argument === null || argument === void 0 ? void 0 : argument.type) === "LogicalExpression")
                    if (argument.operator === "&&") {
                        const sourceCode = context.getSourceCode();
                        context.report({
                            node: argument,
                            messageId: "noConditionalReturn",
                            fix: (fixer) => {
                                const { left: test, right: consequent } = argument;
                                return fixer.replaceText(argument, `<Show when={${sourceCode.getText(test)}}>${putIntoJSX(consequent)}</Show>`);
                            },
                        });
                    }
                    else {
                        context.report({
                            node: argument,
                            messageId: "noConditionalReturn",
                        });
                    }
            }
            functionStack.pop();
        };
        return {
            FunctionDeclaration: onFunctionEnter,
            FunctionExpression: onFunctionEnter,
            ArrowFunctionExpression: onFunctionEnter,
            "FunctionDeclaration:exit": onFunctionExit,
            "FunctionExpression:exit": onFunctionExit,
            "ArrowFunctionExpression:exit": onFunctionExit,
            JSXElement() {
                if (functionStack.length) {
                    currentFunction().isComponent = true;
                }
            },
            JSXFragment() {
                if (functionStack.length) {
                    currentFunction().isComponent = true;
                }
            },
            ReturnStatement(node) {
                if (functionStack.length && node !== currentFunction().lastReturn) {
                    currentFunction().earlyReturns.push(node);
                }
            },
        };
    },
};
exports.default = rule;
