"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@typescript-eslint/utils");
const utils_2 = require("../utils");
const { findVariable, getFunctionHeadLocation } = utils_1.ASTUtils;
class ScopeStackItem {
    constructor(node) {
        this.trackedScopes = [];
        this.unnamedDerivedSignals = new Set();
        this.hasJSX = false;
        this.node = node;
    }
}
class ScopeStack extends Array {
    constructor() {
        super(...arguments);
        this.currentScope = () => this[this.length - 1];
        this.parentScope = () => this[this.length - 2];
        this.syncCallbacks = new Set();
        this.findDeepestDeclarationScope = (a, b) => {
            if (a === b)
                return a;
            for (let i = this.length - 1; i >= 0; i -= 1) {
                const { node } = this[i];
                if (a === node || b === node) {
                    return node;
                }
            }
            throw new Error("This should never happen");
        };
        this.signals = [];
        this.props = [];
    }
    pushSignal(variable, declarationScope = this.currentScope().node) {
        this.signals.push({
            references: variable.references.filter((reference) => !reference.init),
            variable,
            declarationScope,
        });
    }
    pushUniqueSignal(variable, declarationScope) {
        const foundSignal = this.signals.find((s) => s.variable === variable);
        if (!foundSignal) {
            this.pushSignal(variable, declarationScope);
        }
        else {
            foundSignal.declarationScope = this.findDeepestDeclarationScope(foundSignal.declarationScope, declarationScope);
        }
    }
    pushProps(variable, declarationScope = this.currentScope().node) {
        this.props.push({
            references: variable.references.filter((reference) => !reference.init),
            variable,
            declarationScope,
        });
    }
    *consumeSignalReferencesInScope() {
        yield* this.consumeReferencesInScope(this.signals);
        this.signals = this.signals.filter((variable) => variable.references.length !== 0);
    }
    *consumePropsReferencesInScope() {
        yield* this.consumeReferencesInScope(this.props);
        this.props = this.props.filter((variable) => variable.references.length !== 0);
    }
    *consumeReferencesInScope(variables) {
        for (const variable of variables) {
            const { references } = variable;
            const inScope = [], notInScope = [];
            references.forEach((reference) => {
                if (this.isReferenceInCurrentScope(reference)) {
                    inScope.push(reference);
                }
                else {
                    notInScope.push(reference);
                }
            });
            yield* inScope.map((reference) => ({
                reference,
                declarationScope: variable.declarationScope,
            }));
            variable.references = notInScope;
        }
    }
    isReferenceInCurrentScope(reference) {
        let parentFunction = (0, utils_2.findParent)(reference.identifier, utils_2.isProgramOrFunctionNode);
        while ((0, utils_2.isFunctionNode)(parentFunction) && this.syncCallbacks.has(parentFunction)) {
            parentFunction = (0, utils_2.findParent)(parentFunction, utils_2.isProgramOrFunctionNode);
        }
        return parentFunction === this.currentScope().node;
    }
}
const getNthDestructuredVar = (id, n, scope) => {
    if ((id === null || id === void 0 ? void 0 : id.type) === "ArrayPattern") {
        const el = id.elements[n];
        if ((el === null || el === void 0 ? void 0 : el.type) === "Identifier") {
            return findVariable(scope, el.name);
        }
    }
    return null;
};
const getReturnedVar = (id, scope) => {
    if (id.type === "Identifier") {
        return findVariable(scope, id.name);
    }
    return null;
};
const rule = {
    meta: {
        type: "problem",
        docs: {
            recommended: "warn",
            description: "Enforce that reactive expressions (props, signals, memos, etc.) are only used in tracked scopes; otherwise, they won't update the view as expected.",
            url: "https://github.com/solidjs-community/eslint-plugin-solid/blob/main/docs/reactivity.md",
        },
        schema: [],
        messages: {
            noWrite: "The reactive variable '{{name}}' should not be reassigned or altered directly.",
            untrackedReactive: "The reactive variable '{{name}}' should be used within JSX, a tracked scope (like createEffect), or inside an event handler function.",
            expectedFunctionGotExpression: "The reactive variable '{{name}}' should be wrapped in a function for reactivity. This includes event handler bindings on native elements, which are not reactive like other JSX props.",
            badSignal: "The reactive variable '{{name}}' should be called as a function when used in {{where}}.",
            badUnnamedDerivedSignal: "This function should be passed to a tracked scope (like createEffect) or an event handler because it contains reactivity.",
            shouldDestructure: "For proper analysis, array destructuring should be used to capture the {{nth}}result of this function call.",
            shouldAssign: "For proper analysis, a variable should be used to capture the result of this function call.",
            noAsyncTrackedScope: "This tracked scope should not be async. Solid's reactivity only tracks synchronously.",
        },
    },
    create(context) {
        const warnShouldDestructure = (node, nth) => context.report({
            node,
            messageId: "shouldDestructure",
            data: nth ? { nth: nth + " " } : undefined,
        });
        const warnShouldAssign = (node) => context.report({ node, messageId: "shouldAssign" });
        const sourceCode = context.getSourceCode();
        const scopeStack = new ScopeStack();
        const { currentScope, parentScope } = scopeStack;
        const { matchImport, handleImportDeclaration } = (0, utils_2.trackImports)();
        const onFunctionEnter = (node) => {
            if ((0, utils_2.isFunctionNode)(node) && scopeStack.syncCallbacks.has(node)) {
                return;
            }
            scopeStack.push(new ScopeStackItem(node));
        };
        const matchTrackedScope = (trackedScope, node) => {
            switch (trackedScope.expect) {
                case "function":
                case "called-function":
                    return node === trackedScope.node;
                case "expression":
                    return Boolean((0, utils_2.findInScope)(node, currentScope().node, (node) => node === trackedScope.node));
            }
        };
        const handleTrackedScopes = (identifier, declarationScope) => {
            var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
            const currentScopeNode = currentScope().node;
            if (!currentScope().trackedScopes.find((trackedScope) => matchTrackedScope(trackedScope, identifier))) {
                const matchedExpression = currentScope().trackedScopes.find((trackedScope) => matchTrackedScope(Object.assign(Object.assign({}, trackedScope), { expect: "expression" }), identifier));
                if (declarationScope === currentScopeNode) {
                    let parentMemberExpression = null;
                    if (((_a = identifier.parent) === null || _a === void 0 ? void 0 : _a.type) === "MemberExpression") {
                        parentMemberExpression = identifier.parent;
                        while (((_b = parentMemberExpression.parent) === null || _b === void 0 ? void 0 : _b.type) === "MemberExpression") {
                            parentMemberExpression = parentMemberExpression.parent;
                        }
                    }
                    const parentCallExpression = ((_c = identifier.parent) === null || _c === void 0 ? void 0 : _c.type) === "CallExpression" ? identifier.parent : null;
                    context.report({
                        node: (_d = parentMemberExpression !== null && parentMemberExpression !== void 0 ? parentMemberExpression : parentCallExpression) !== null && _d !== void 0 ? _d : identifier,
                        messageId: matchedExpression ? "expectedFunctionGotExpression" : "untrackedReactive",
                        data: {
                            name: parentMemberExpression
                                ? sourceCode.getText(parentMemberExpression)
                                : identifier.name,
                        },
                    });
                }
                else {
                    if (!parentScope() || !(0, utils_2.isFunctionNode)(currentScopeNode)) {
                        throw new Error("this shouldn't happen!");
                    }
                    const pushUnnamedDerivedSignal = () => { var _a; var _b; return ((_a = (_b = parentScope()).unnamedDerivedSignals) !== null && _a !== void 0 ? _a : (_b.unnamedDerivedSignals = new Set())).add(currentScopeNode); };
                    if (currentScopeNode.type === "FunctionDeclaration") {
                        const functionVariable = (_f = (_e = sourceCode.scopeManager) === null || _e === void 0 ? void 0 : _e.getDeclaredVariables(currentScopeNode)) === null || _f === void 0 ? void 0 : _f[0];
                        if (functionVariable) {
                            scopeStack.pushUniqueSignal(functionVariable, declarationScope);
                        }
                        else {
                            pushUnnamedDerivedSignal();
                        }
                    }
                    else if (((_g = currentScopeNode.parent) === null || _g === void 0 ? void 0 : _g.type) === "VariableDeclarator") {
                        const declarator = currentScopeNode.parent;
                        const functionVariable = (_j = (_h = sourceCode.scopeManager) === null || _h === void 0 ? void 0 : _h.getDeclaredVariables(declarator)) === null || _j === void 0 ? void 0 : _j[0];
                        if (functionVariable) {
                            scopeStack.pushUniqueSignal(functionVariable, declarationScope);
                        }
                        else {
                            pushUnnamedDerivedSignal();
                        }
                    }
                    else if (((_k = currentScopeNode.parent) === null || _k === void 0 ? void 0 : _k.type) === "Property") {
                    }
                    else {
                        pushUnnamedDerivedSignal();
                    }
                }
            }
        };
        const onFunctionExit = (currentScopeNode) => {
            var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s;
            if ((0, utils_2.isFunctionNode)(currentScopeNode) && scopeStack.syncCallbacks.has(currentScopeNode)) {
                return;
            }
            if ((0, utils_2.isFunctionNode)(currentScopeNode) && currentScopeNode.params.length === 1) {
                const paramsNode = currentScopeNode.params[0];
                if ((paramsNode === null || paramsNode === void 0 ? void 0 : paramsNode.type) === "Identifier" &&
                    ((currentScope().hasJSX &&
                        (currentScopeNode.type !== "FunctionDeclaration" ||
                            !((_b = (_a = currentScopeNode.id) === null || _a === void 0 ? void 0 : _a.name) === null || _b === void 0 ? void 0 : _b.match(/^[a-z]/)))) ||
                        (0, utils_2.isPropsByName)(paramsNode.name)) &&
                    ((_c = currentScopeNode.parent) === null || _c === void 0 ? void 0 : _c.type) !== "JSXExpressionContainer" &&
                    ((_d = currentScopeNode.parent) === null || _d === void 0 ? void 0 : _d.type) !== "TemplateLiteral") {
                    const propsParam = findVariable(context.getScope(), paramsNode);
                    if (propsParam) {
                        scopeStack.pushProps(propsParam);
                    }
                }
            }
            for (const { reference, declarationScope } of scopeStack.consumeSignalReferencesInScope()) {
                const identifier = reference.identifier;
                if (reference.isWrite()) {
                    context.report({
                        node: identifier,
                        messageId: "noWrite",
                        data: {
                            name: identifier.name,
                        },
                    });
                }
                else if (identifier.type === "Identifier") {
                    const reportBadSignal = (where) => context.report({
                        node: identifier,
                        messageId: "badSignal",
                        data: { name: identifier.name, where },
                    });
                    if (((_e = identifier.parent) === null || _e === void 0 ? void 0 : _e.type) === "CallExpression" ||
                        (((_f = identifier.parent) === null || _f === void 0 ? void 0 : _f.type) === "ArrayExpression" &&
                            ((_g = identifier.parent.parent) === null || _g === void 0 ? void 0 : _g.type) === "CallExpression")) {
                        handleTrackedScopes(identifier, declarationScope);
                    }
                    else if (((_h = identifier.parent) === null || _h === void 0 ? void 0 : _h.type) === "TemplateLiteral") {
                        reportBadSignal("template literals");
                    }
                    else if (((_j = identifier.parent) === null || _j === void 0 ? void 0 : _j.type) === "BinaryExpression" &&
                        [
                            "<",
                            "<=",
                            ">",
                            ">=",
                            "<<",
                            ">>",
                            ">>>",
                            "+",
                            "-",
                            "*",
                            "/",
                            "%",
                            "**",
                            "|",
                            "^",
                            "&",
                            "in",
                        ].includes(identifier.parent.operator)) {
                        reportBadSignal("arithmetic or comparisons");
                    }
                    else if (((_k = identifier.parent) === null || _k === void 0 ? void 0 : _k.type) === "UnaryExpression" &&
                        ["-", "+", "~"].includes(identifier.parent.operator)) {
                        reportBadSignal("unary expressions");
                    }
                    else if (((_l = identifier.parent) === null || _l === void 0 ? void 0 : _l.type) === "MemberExpression" &&
                        identifier.parent.computed &&
                        identifier.parent.property === identifier) {
                        reportBadSignal("property accesses");
                    }
                    else if (((_m = identifier.parent) === null || _m === void 0 ? void 0 : _m.type) === "JSXExpressionContainer" &&
                        !currentScope().trackedScopes.find((trackedScope) => trackedScope.node === identifier &&
                            (trackedScope.expect === "function" || trackedScope.expect === "called-function"))) {
                        const elementOrAttribute = identifier.parent.parent;
                        if ((elementOrAttribute === null || elementOrAttribute === void 0 ? void 0 : elementOrAttribute.type) === "JSXElement" ||
                            ((elementOrAttribute === null || elementOrAttribute === void 0 ? void 0 : elementOrAttribute.type) === "JSXAttribute" &&
                                ((_o = elementOrAttribute.parent) === null || _o === void 0 ? void 0 : _o.type) === "JSXOpeningElement" &&
                                elementOrAttribute.parent.name.type === "JSXIdentifier" &&
                                (0, utils_2.isDOMElementName)(elementOrAttribute.parent.name.name))) {
                            reportBadSignal("JSX");
                        }
                    }
                }
            }
            for (const { reference, declarationScope } of scopeStack.consumePropsReferencesInScope()) {
                const identifier = reference.identifier;
                if (reference.isWrite()) {
                    context.report({
                        node: identifier,
                        messageId: "noWrite",
                        data: {
                            name: identifier.name,
                        },
                    });
                }
                else if (((_p = identifier.parent) === null || _p === void 0 ? void 0 : _p.type) === "MemberExpression" &&
                    identifier.parent.object === identifier) {
                    const { parent } = identifier;
                    if (((_q = parent.parent) === null || _q === void 0 ? void 0 : _q.type) === "AssignmentExpression" && parent.parent.left === parent) {
                        context.report({
                            node: identifier,
                            messageId: "noWrite",
                            data: {
                                name: identifier.name,
                            },
                        });
                    }
                    else if (parent.property.type === "Identifier" &&
                        /^(?:initial|default)[A-Z]/.test(parent.property.name)) {
                    }
                    else {
                        handleTrackedScopes(identifier, declarationScope);
                    }
                }
                else if (((_r = identifier.parent) === null || _r === void 0 ? void 0 : _r.type) === "AssignmentExpression" ||
                    ((_s = identifier.parent) === null || _s === void 0 ? void 0 : _s.type) === "VariableDeclarator") {
                    context.report({
                        node: identifier,
                        messageId: "untrackedReactive",
                        data: { name: identifier.name },
                    });
                }
            }
            const { unnamedDerivedSignals } = currentScope();
            if (unnamedDerivedSignals) {
                for (const node of unnamedDerivedSignals) {
                    if (!currentScope().trackedScopes.find((trackedScope) => trackedScope.node === node &&
                        (trackedScope.expect === "function" || trackedScope.expect === "called-function"))) {
                        context.report({
                            loc: getFunctionHeadLocation(node, sourceCode),
                            messageId: "badUnnamedDerivedSignal",
                        });
                    }
                }
            }
            scopeStack.pop();
        };
        const checkForSyncCallbacks = (node) => {
            var _a, _b;
            if (node.arguments.length === 1 &&
                (0, utils_2.isFunctionNode)(node.arguments[0]) &&
                !node.arguments[0].async) {
                if (node.callee.type === "Identifier" &&
                    matchImport(["batch", "produce"], node.callee.name)) {
                    scopeStack.syncCallbacks.add(node.arguments[0]);
                }
                else if (node.callee.type === "MemberExpression" &&
                    !node.callee.computed &&
                    node.callee.object.type !== "ObjectExpression" &&
                    /^(?:forEach|map|flatMap|reduce|reduceRight|find|findIndex|filter|every|some)$/.test(node.callee.property.name)) {
                    scopeStack.syncCallbacks.add(node.arguments[0]);
                }
            }
            if (node.callee.type === "Identifier") {
                if (matchImport(["createSignal", "createStore"], node.callee.name) &&
                    ((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === "VariableDeclarator") {
                    const setter = getNthDestructuredVar(node.parent.id, 1, context.getScope());
                    if (setter) {
                        for (const reference of setter.references) {
                            const { identifier } = reference;
                            if (!reference.init &&
                                reference.isRead() &&
                                ((_b = identifier.parent) === null || _b === void 0 ? void 0 : _b.type) === "CallExpression") {
                                for (const arg of identifier.parent.arguments) {
                                    if ((0, utils_2.isFunctionNode)(arg) && !arg.async) {
                                        scopeStack.syncCallbacks.add(arg);
                                    }
                                }
                            }
                        }
                    }
                }
                else if (matchImport(["mapArray", "indexArray"], node.callee.name)) {
                    const arg1 = node.arguments[1];
                    if ((0, utils_2.isFunctionNode)(arg1)) {
                        scopeStack.syncCallbacks.add(arg1);
                    }
                }
            }
            if ((0, utils_2.isFunctionNode)(node.callee)) {
                scopeStack.syncCallbacks.add(node.callee);
            }
        };
        const checkForReactiveAssignment = (id, init) => {
            if (init.type === "CallExpression" && init.callee.type === "Identifier") {
                const { callee } = init;
                if (matchImport(["createSignal", "useTransition"], callee.name)) {
                    const signal = id && getNthDestructuredVar(id, 0, context.getScope());
                    if (signal) {
                        scopeStack.pushSignal(signal, currentScope().node);
                    }
                    else {
                        warnShouldDestructure(id !== null && id !== void 0 ? id : init, "first");
                    }
                }
                else if (matchImport(["createMemo", "createSelector"], callee.name)) {
                    const memo = id && getReturnedVar(id, context.getScope());
                    if (memo) {
                        scopeStack.pushSignal(memo, currentScope().node);
                    }
                    else {
                        warnShouldAssign(id !== null && id !== void 0 ? id : init);
                    }
                }
                else if (matchImport("createStore", callee.name)) {
                    const store = id && getNthDestructuredVar(id, 0, context.getScope());
                    if (store) {
                        scopeStack.pushProps(store, currentScope().node);
                    }
                    else {
                        warnShouldDestructure(id !== null && id !== void 0 ? id : init, "first");
                    }
                }
                else if (matchImport("mergeProps", callee.name)) {
                    const merged = id && getReturnedVar(id, context.getScope());
                    if (merged) {
                        scopeStack.pushProps(merged, currentScope().node);
                    }
                    else {
                        warnShouldAssign(id !== null && id !== void 0 ? id : init);
                    }
                }
                else if (matchImport("splitProps", callee.name)) {
                    if ((id === null || id === void 0 ? void 0 : id.type) === "ArrayPattern") {
                        const vars = id.elements
                            .map((_, i) => getNthDestructuredVar(id, i, context.getScope()))
                            .filter(Boolean);
                        if (vars.length === 0) {
                            warnShouldDestructure(id);
                        }
                        else {
                            vars.forEach((variable) => {
                                scopeStack.pushProps(variable, currentScope().node);
                            });
                        }
                    }
                    else {
                        const vars = id && getReturnedVar(id, context.getScope());
                        if (vars) {
                            scopeStack.pushProps(vars, currentScope().node);
                        }
                    }
                }
                else if (matchImport("createResource", callee.name)) {
                    const resourceReturn = id && getNthDestructuredVar(id, 0, context.getScope());
                    if (resourceReturn) {
                        scopeStack.pushProps(resourceReturn, currentScope().node);
                    }
                }
                else if (matchImport("createMutable", callee.name)) {
                    const mutable = id && getReturnedVar(id, context.getScope());
                    if (mutable) {
                        scopeStack.pushProps(mutable, currentScope().node);
                    }
                }
                else if (matchImport("mapArray", callee.name)) {
                    const arg1 = init.arguments[1];
                    if ((0, utils_2.isFunctionNode)(arg1) &&
                        arg1.params.length >= 2 &&
                        arg1.params[1].type === "Identifier") {
                        const indexSignal = findVariable(context.getScope(), arg1.params[1]);
                        if (indexSignal) {
                            scopeStack.pushSignal(indexSignal);
                        }
                    }
                }
                else if (matchImport("indexArray", callee.name)) {
                    const arg1 = init.arguments[1];
                    if ((0, utils_2.isFunctionNode)(arg1) &&
                        arg1.params.length >= 1 &&
                        arg1.params[0].type === "Identifier") {
                        const valueSignal = findVariable(context.getScope(), arg1.params[0]);
                        if (valueSignal) {
                            scopeStack.pushSignal(valueSignal);
                        }
                    }
                }
            }
        };
        const checkForTrackedScopes = (node) => {
            var _a, _b, _c, _d, _e, _f, _g, _h;
            const pushTrackedScope = (node, expect) => {
                currentScope().trackedScopes.push({ node, expect });
                if (expect !== "called-function" && (0, utils_2.isFunctionNode)(node) && node.async) {
                    context.report({
                        node,
                        messageId: "noAsyncTrackedScope",
                    });
                }
            };
            if (node.type === "JSXExpressionContainer") {
                if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === "JSXAttribute" &&
                    /^on[:A-Z]/.test(sourceCode.getText(node.parent.name)) &&
                    ((_b = node.parent.parent) === null || _b === void 0 ? void 0 : _b.type) === "JSXOpeningElement" &&
                    node.parent.parent.name.type === "JSXIdentifier" &&
                    (0, utils_2.isDOMElementName)(node.parent.parent.name.name)) {
                    pushTrackedScope(node.expression, "called-function");
                }
                else if (((_c = node.parent) === null || _c === void 0 ? void 0 : _c.type) === "JSXAttribute" &&
                    node.parent.name.name === "value" &&
                    ((_d = node.parent.parent) === null || _d === void 0 ? void 0 : _d.type) === "JSXOpeningElement" &&
                    ((node.parent.parent.name.type === "JSXIdentifier" &&
                        node.parent.parent.name.name.endsWith("Provider")) ||
                        (node.parent.parent.name.type === "JSXMemberExpression" &&
                            node.parent.parent.name.property.name === "Provider"))) {
                }
                else if (((_e = node.parent) === null || _e === void 0 ? void 0 : _e.type) === "JSXAttribute" &&
                    node.parent.name.name === "ref" &&
                    (0, utils_2.isFunctionNode)(node.expression)) {
                    pushTrackedScope(node.expression, "called-function");
                }
                else {
                    pushTrackedScope(node.expression, "expression");
                }
            }
            else if (node.type === "JSXSpreadAttribute") {
                pushTrackedScope(node.argument, "expression");
            }
            else if (node.type === "CallExpression") {
                if (node.callee.type === "Identifier") {
                    const { callee, arguments: { 0: arg0, 1: arg1 }, } = node;
                    if (matchImport([
                        "createMemo",
                        "children",
                        "createEffect",
                        "createRenderEffect",
                        "createDeferred",
                        "createComputed",
                        "createSelector",
                        "untrack",
                        "mapArray",
                        "indexArray",
                    ], callee.name) ||
                        (matchImport("createResource", callee.name) && node.arguments.length >= 2)) {
                        pushTrackedScope(arg0, "function");
                    }
                    else if (matchImport(["onMount", "onCleanup", "onError"], callee.name) ||
                        [
                            "setInterval",
                            "setTimeout",
                            "setImmediate",
                            "requestAnimationFrame",
                            "requestIdleCallback",
                        ].includes(callee.name)) {
                        pushTrackedScope(arg0, "called-function");
                    }
                    else if (matchImport("createMutable", callee.name) && arg0) {
                        pushTrackedScope(arg0, "expression");
                    }
                    else if (matchImport("on", callee.name)) {
                        if (arg0) {
                            if (arg0.type === "ArrayExpression") {
                                arg0.elements.forEach((element) => {
                                    pushTrackedScope(element, "function");
                                });
                            }
                            else {
                                pushTrackedScope(arg0, "function");
                            }
                        }
                        if (arg1) {
                            pushTrackedScope(arg1, "called-function");
                        }
                    }
                    else if (matchImport("createStore", callee.name) && (arg0 === null || arg0 === void 0 ? void 0 : arg0.type) === "ObjectExpression") {
                        for (const property of arg0.properties) {
                            if (property.type === "Property" &&
                                property.kind === "get" &&
                                (0, utils_2.isFunctionNode)(property.value)) {
                                pushTrackedScope(property.value, "function");
                            }
                        }
                    }
                    else if (matchImport("runWithOwner", callee.name)) {
                        if (arg1) {
                            let isTrackedScope = true;
                            const owner = arg0.type === "Identifier" && findVariable(context.getScope(), arg0);
                            if (owner) {
                                const decl = owner.defs[0];
                                if (decl &&
                                    decl.node.type === "VariableDeclarator" &&
                                    ((_f = decl.node.init) === null || _f === void 0 ? void 0 : _f.type) === "CallExpression" &&
                                    decl.node.init.callee.type === "Identifier" &&
                                    matchImport("getOwner", decl.node.init.callee.name)) {
                                    const ownerFunction = (0, utils_2.findParent)(decl.node, utils_2.isProgramOrFunctionNode);
                                    const scopeStackIndex = scopeStack.findIndex(({ node }) => ownerFunction === node);
                                    if ((scopeStackIndex >= 1 &&
                                        !scopeStack[scopeStackIndex - 1].trackedScopes.some((trackedScope) => trackedScope.expect === "function" && trackedScope.node === ownerFunction)) ||
                                        scopeStackIndex === 0) {
                                        isTrackedScope = false;
                                    }
                                }
                            }
                            if (isTrackedScope) {
                                pushTrackedScope(arg1, "function");
                            }
                        }
                    }
                    else if (/^(?:use|create)[A-Z]/.test(callee.name)) {
                        for (const arg of node.arguments) {
                            if ((0, utils_2.isFunctionNode)(arg)) {
                                pushTrackedScope(arg, "called-function");
                            }
                            else if (arg.type === "Identifier" ||
                                arg.type === "ObjectExpression" ||
                                arg.type === "ArrayExpression") {
                                pushTrackedScope(arg, "expression");
                            }
                        }
                    }
                }
                else if (node.callee.type === "MemberExpression") {
                    const { property } = node.callee;
                    if (property.type === "Identifier" &&
                        property.name === "addEventListener" &&
                        node.arguments.length >= 2) {
                        pushTrackedScope(node.arguments[1], "called-function");
                    }
                    else if (property.type === "Identifier" && /^(?:use|create)[A-Z]/.test(property.name)) {
                        for (const arg of node.arguments) {
                            if ((0, utils_2.isFunctionNode)(arg)) {
                                pushTrackedScope(arg, "called-function");
                            }
                            else if (arg.type === "Identifier" ||
                                arg.type === "ObjectExpression" ||
                                arg.type === "ArrayExpression") {
                                pushTrackedScope(arg, "expression");
                            }
                        }
                    }
                }
            }
            else if (node.type === "VariableDeclarator") {
                if (((_g = node.init) === null || _g === void 0 ? void 0 : _g.type) === "CallExpression" && node.init.callee.type === "Identifier") {
                    if (matchImport(["createReactive", "createReaction"], node.init.callee.name)) {
                        const track = getReturnedVar(node.id, context.getScope());
                        if (track) {
                            for (const reference of track.references) {
                                if (!reference.init &&
                                    reference.isReadOnly() &&
                                    ((_h = reference.identifier.parent) === null || _h === void 0 ? void 0 : _h.type) === "CallExpression" &&
                                    reference.identifier.parent.callee === reference.identifier) {
                                    const arg0 = reference.identifier.parent.arguments[0];
                                    arg0 && pushTrackedScope(arg0, "function");
                                }
                            }
                        }
                        if ((0, utils_2.isFunctionNode)(node.init.arguments[0])) {
                            pushTrackedScope(node.init.arguments[0], "called-function");
                        }
                    }
                }
            }
            else if (node.type === "AssignmentExpression") {
                if (node.left.type === "MemberExpression" &&
                    node.left.property.type === "Identifier" &&
                    (0, utils_2.isFunctionNode)(node.right) &&
                    /^on[a-z]+$/.test(node.left.property.name)) {
                    pushTrackedScope(node.right, "called-function");
                }
            }
            else if (node.type === "TaggedTemplateExpression") {
                for (const expression of node.quasi.expressions) {
                    if ((0, utils_2.isFunctionNode)(expression)) {
                        pushTrackedScope(expression, "called-function");
                        for (const param of expression.params) {
                            if (param.type === "Identifier" && (0, utils_2.isPropsByName)(param.name)) {
                                const variable = findVariable(context.getScope(), param);
                                if (variable)
                                    scopeStack.pushProps(variable, currentScope().node);
                            }
                        }
                    }
                }
            }
        };
        return {
            ImportDeclaration: handleImportDeclaration,
            JSXExpressionContainer(node) {
                checkForTrackedScopes(node);
            },
            JSXSpreadAttribute(node) {
                checkForTrackedScopes(node);
            },
            CallExpression(node) {
                var _a, _b;
                checkForTrackedScopes(node);
                checkForSyncCallbacks(node);
                if (((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) !== "AssignmentExpression" &&
                    ((_b = node.parent) === null || _b === void 0 ? void 0 : _b.type) !== "VariableDeclarator") {
                    checkForReactiveAssignment(null, node);
                }
            },
            VariableDeclarator(node) {
                if (node.init) {
                    checkForReactiveAssignment(node.id, node.init);
                    checkForTrackedScopes(node);
                }
            },
            AssignmentExpression(node) {
                if (node.left.type !== "MemberExpression") {
                    checkForReactiveAssignment(node.left, node.right);
                }
                checkForTrackedScopes(node);
            },
            TaggedTemplateExpression(node) {
                checkForTrackedScopes(node);
            },
            "JSXElement > JSXExpressionContainer > :function"(node) {
                var _a, _b;
                if ((0, utils_2.isFunctionNode)(node) &&
                    ((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === "JSXExpressionContainer" &&
                    ((_b = node.parent.parent) === null || _b === void 0 ? void 0 : _b.type) === "JSXElement") {
                    const element = node.parent.parent;
                    if (element.openingElement.name.type === "JSXIdentifier") {
                        const tagName = element.openingElement.name.name;
                        if (matchImport("For", tagName) &&
                            node.params.length === 2 &&
                            node.params[1].type === "Identifier") {
                            const index = findVariable(context.getScope(), node.params[1]);
                            if (index) {
                                scopeStack.pushSignal(index, currentScope().node);
                            }
                        }
                        else if (matchImport("Index", tagName) &&
                            node.params.length >= 1 &&
                            node.params[0].type === "Identifier") {
                            const item = findVariable(context.getScope(), node.params[0]);
                            if (item) {
                                scopeStack.pushSignal(item, currentScope().node);
                            }
                        }
                    }
                }
            },
            FunctionExpression: onFunctionEnter,
            ArrowFunctionExpression: onFunctionEnter,
            FunctionDeclaration: onFunctionEnter,
            Program: onFunctionEnter,
            "FunctionExpression:exit": onFunctionExit,
            "ArrowFunctionExpression:exit": onFunctionExit,
            "FunctionDeclaration:exit": onFunctionExit,
            "Program:exit": onFunctionExit,
            JSXElement() {
                if (scopeStack.length) {
                    currentScope().hasJSX = true;
                }
            },
            JSXFragment() {
                if (scopeStack.length) {
                    currentScope().hasJSX = true;
                }
            },
        };
    },
};
exports.default = rule;
